using System;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class Force : MonoBehaviour
    {
        public new string name;
        public Rigidbody rb;
        public Func<Vector3> GetDirection;

        public ForceParam amplitude;
        public ForceParam period;
        public ForceParam periodRatio;
        public ForceParam quickness;

        public JSONStorableBool enabledJ => control.enabled;
        public JSONStorableBool applyReturn = new JSONStorableBool("Apply On Return", true);
        public JSONStorableBool constant = new JSONStorableBool("Constant Force", false);

        public float timer;
        public float flip = 1f;
        protected Vector3 targetForce;
        protected Vector3 currentForce;
        public bool shutDown;

        // public float offset;
        public float offsetTarget;
        private float currentOffset;
        private float offsetQuickness;
        public bool offsetAtTarget;

        public ForceControl control;
        public ForceSync sync;
        protected StringBuilder sb = new StringBuilder();

        public List<SyncedForceGroup> forceGroups = new List<SyncedForceGroup>();
        public Dictionary<SyncedForceGroup, int> priorities = new Dictionary<SyncedForceGroup, int>();

        private bool initialized;

        public Force Init(string name, Rigidbody rb, Func<Vector3> getDirection)
        {
            this.name = name;
            this.rb = rb;
            GetDirection = getDirection;
            amplitude = new ForceParam("amplitude",0f, 0f);
            period = new ForceParam("period", .5f, .25f);
            periodRatio = new ForceParam("periodRatio", .5f, .2f);
            quickness = new ForceParam("quickness", 4f, 3f);
            period.current = period.mean.val;
            periodRatio.current = periodRatio.mean.val;
            enabled = false;
            control = new ForceControl(this);
            constant.setCallbackFunction += val => flip = 1f;
            ForceToggle.AddForce(this);
            initialized = true;
            return this;
        }

        private void OnEnable()
        {
            flip = 0f;
            timer = 0f;
            amplitude.Reset();
            // period.Reset();
            // periodRatio.Reset();
            // quickness.Reset();
            currentForce = targetForce = Vector3.zero;
            shutDown = false;
        }

        private void OnDisable()
        {
            currentOffset = 0f;
            control.info.val = "Inactive";
        }

        public void ShutDown(float quickness = 0f)
        {
            if (!enabled) return;
            amplitude.target = 0f;
            if (quickness > 0f)
            {
                amplitude.transitionQuickness = quickness;
                offsetQuickness = quickness*.5f;
            }
            offsetTarget = 0f;
            offsetAtTarget = false;
            shutDown = true;
            
            if (amplitude.atTarget) enabled = false;
            for (var index = 0; index < forceGroups.Count; index++)
            {
                forceGroups[index].MarkDisabled(this);
            }
        }
        
        public void ShutDownImmediate()
        {
            if (!enabled) return;
            amplitude.target = amplitude.current = 0f;
            shutDown = false;
            offsetTarget = currentOffset = 0f;
            enabled = false;
            for (var index = 0; index < forceGroups.Count; index++)
            {
                forceGroups[index].MarkDisabled(this);
            }
        }

        public void Enable()
        {
            if(!initialized || ! enabledJ.val) return;
            shutDown = false;
            amplitude.GetNewTarget();
            offsetTarget = control.offset.val;
            offsetQuickness = control.offsetQuickness.val;
            offsetAtTarget = false;
            enabled = true;
            for (var index = 0; index < forceGroups.Count; index++)
            {
                forceGroups[index].MarkEnabled(this);
            }
        }
        
        public void SetActive(bool val)
        {
            if(val) Enable();
            else ShutDown(4f);
        }
        
        public void SetActiveImmediate(bool val)
        {
            if(val) Enable();
            else ShutDownImmediate();
        }

        public ForceSync AddSync(Force driver = null)
        {
            if (sync != null) return sync;
            sync = new ForceSync(this, driver);
            return sync;
        }

        public void UpdateParams()
        {
            if(!shutDown) amplitude.Update();
            else amplitude.LerpToTarget();
            period.Update();
            periodRatio.Update();
            quickness.Update();
        }

        public void GetPhase()
        {
            timer -= Time.fixedDeltaTime;
            if (timer < 0.0f) {
                if ((flip > 0f && periodRatio.current != 1f) || periodRatio.current == 0f) {
                    flip = applyReturn.val ? -1f : 0f;
                    timer = period.current * (1f - periodRatio.current);
                } else {
                    flip = 1f;
                    timer = period.current * periodRatio.current;
                }
            }
        }

        public virtual void FixedUpdate() {
            try
            {
                if (!offsetAtTarget)
                {
                    currentOffset = Mathf.Lerp(currentOffset, offsetTarget, offsetQuickness*Time.fixedDeltaTime);
                    if (Mathf.Abs(currentOffset - offsetTarget) < 1f)
                    {
                        currentOffset = offsetTarget;
                        offsetAtTarget = true;
                    }
                }
                // currentOffset.Print();
                if(sync != null && sync.enabled && sync.driver.enabled)
                {
                    if(!shutDown) sync.UpdateParams();
                    else amplitude.LerpToTarget();
                    sync.GetPhase();
                    targetForce = (flip * amplitude.current + currentOffset) * GetDirection();
                    currentForce = Vector3.Lerp(currentForce, targetForce, Time.fixedDeltaTime * quickness.current/period.current);
                }
                else
                {
                    if(!constant.val)
                    {
                        UpdateParams();
                        GetPhase();
                        targetForce = (flip * amplitude.current + currentOffset) * GetDirection();
                        currentForce = Vector3.Lerp(currentForce, targetForce, Time.fixedDeltaTime * quickness.current/period.current);
                    }
                    else
                    {
                        if(!shutDown) amplitude.Update();
                        else amplitude.LerpToTarget();
                        currentForce = (amplitude.current + currentOffset) * GetDirection();
                    }
                }

                if (rb && !SuperController.singleton.freezeAnimation) AddForce();
                if (shutDown && amplitude.atTarget && offsetAtTarget)
                {
                    enabled = false;
                }
                if (control.UIOpen)
                {
                    if(!constant.val)
                    {
                        sb = sb.Clear(87)
                            .AppendLine($"<b>{rb.name}</b>" + (sync != null && sync.enabled
                                ? $" (<color=#ff0000>Synced</color> to {sync.driver.name})"
                                : " (Master)"))
                            .Append("Amplitude: ").AppendFormat("{0:F1}", amplitude.current).Append("\n")
                            .Append("Period: ").AppendFormat("{0:F3}", period.current).Append("\n")
                            .Append("PeriodRatio: ").AppendFormat("{0:F3}", periodRatio.current).Append("\n")
                            .Append("Quickness: ").AppendFormat("{0:F2}", quickness.current).Append("\n")
                            .Append("Offset: ").AppendFormat("{0:F2}", currentOffset).Append("\n");
                    }
                    else
                    {
                        sb = sb.Clear(87)
                            .AppendLine($"<b>{rb.name}</b>" + " (constant)")
                            .Append("Amplitude: ").AppendFormat("{0:F1}", amplitude.current);
                    }
                    control.info.val = sb.ToString();
                }
                // currentForce.Print();
            }
            catch (Exception e) {
                SuperController.LogError($"Exception caught: {name} {e}");
            }
        }

        protected virtual void AddForce()
        {
            rb.AddForce(currentForce, ForceMode.Force);
        }

        public JSONClass Store => control.Store();
        
        public void Load(JSONClass jc) => control.Load(jc);
        
        // public void Load1(JSONClass jc)
        // {
        //     if (!jc.HasKey(name)) return;
        //     control.Load(jc[name].AsObject);
        // }
    }
}